from .scope_config_base import ScopeConfigureBase

class StripToolConfigure(ScopeConfigureBase):
    def __init__(self, params, **kwargs):
        super().__init__(params,
                         show_start=False,
                         show_trigger=False,
                         **kwargs)

    def add_source_acquisition_props(self, children, section):
        children.append({"name": "Sample Mode", "type":"bool", "value": True})
        return children
    
    def parse(self):
        try:
            acquisition = self.assemble_acquisition(self.params["ACQUISITION"])
        except KeyError:
            acquisition = self.assemble_acquisition()

        try:
            display = self.assemble_display(self.params["DISPLAY"])
        except KeyError:
            display = self.assemble_display()
            
        cfg = self.assemble_config()
        statistics = self.assemble_statistics()
        # line up in order
        paramcfg = [acquisition, display, cfg, statistics]

        return paramcfg





# -*- coding: utf-8 -*-

"""
Copyright 2021 UChicago Argonne LLC
 as operator of Argonne National Laboratory

@author: Matic Pogacnik <mpogacnik@anl.gov>
"""

import numpy as np
import pyqtgraph as pg

from .image_definitions import COLOR_MODE_MONO


class ImageProfileWidget(object):
    """
    Part of PVA object viewer utilities, it is responsible to
    display and image profiles.
    """

#########################################
#   Public interface
#########################################

    def __init__(self, canvas_grid_reference):
        super().__init__()

        # Reference to the canvas grid where the image and profiles are displayed.
        self._grid = canvas_grid_reference

        # Internal Numpy array holding image data
        self._data = None
        self._data_color_mode = None

        # Profiles of the data
        self._x_profile = None
        self._y_profile = None

        # Mutex to protect the data variable as set_image_data and plot are
        # called from the different threads
        self._data_mutex = pg.Qt.QtCore.QMutex()

        # Status of the profiles UI
        self._display_profiles = False

        # Save the reference to the image widget
        self._image_widget = self._grid.itemAtPosition(1, 1)

        # Build the X profile graph widget
        self._plot_x_profile = self._grid.itemAtPosition(0, 1)
        self._plot_x_profile.widget().getPlotItem().hideAxis('bottom')
        self._plot_x_profile.widget().getPlotItem().hideAxis('left')

        plot_x_size_policy = pg.Qt.QtGui.QSizePolicy(pg.Qt.QtGui.QSizePolicy.Fixed, pg.Qt.QtGui.QSizePolicy.Fixed)
        plot_x_size_policy.setHeightForWidth(self._plot_x_profile.widget().sizePolicy().hasHeightForWidth())
        self._plot_x_profile.widget().setSizePolicy(plot_x_size_policy)

        self._profile_x_curves = {
            'mono'  : self._plot_x_profile.widget().plot(pen=pg.mkPen(color=(255,255,255), style=pg.Qt.QtCore.Qt.SolidLine)),
            'red'   : self._plot_x_profile.widget().plot(pen=pg.mkPen(color=(255,  0,  0), style=pg.Qt.QtCore.Qt.SolidLine)),
            'green' : self._plot_x_profile.widget().plot(pen=pg.mkPen(color=(  0,255,  0), style=pg.Qt.QtCore.Qt.DashLine)),
            'blue'  : self._plot_x_profile.widget().plot(pen=pg.mkPen(color=(  0,255,255), style=pg.Qt.QtCore.Qt.DashDotLine)),
        }

        # Build the Y profile graph widget
        self._plot_y_profile = self._grid.itemAtPosition(1, 0)
        self._plot_y_profile.widget().getPlotItem().hideAxis('bottom')
        self._plot_y_profile.widget().getPlotItem().hideAxis('left')

        plot_y_size_policy = pg.Qt.QtGui.QSizePolicy(pg.Qt.QtGui.QSizePolicy.Fixed, pg.Qt.QtGui.QSizePolicy.Fixed)
        plot_y_size_policy.setHeightForWidth(self._plot_y_profile.widget().sizePolicy().hasHeightForWidth())
        self._plot_y_profile.widget().setSizePolicy(plot_y_size_policy)

        self._profile_y_curves = {
            'mono'  : self._plot_y_profile.widget().plot(pen=pg.mkPen(color=(255,255,255), style=pg.Qt.QtCore.Qt.SolidLine)),
            'red'   : self._plot_y_profile.widget().plot(pen=pg.mkPen(color=(255,  0,  0), style=pg.Qt.QtCore.Qt.SolidLine)),
            'green' : self._plot_y_profile.widget().plot(pen=pg.mkPen(color=(  0,255,  0), style=pg.Qt.QtCore.Qt.DashLine)),
            'blue'  : self._plot_y_profile.widget().plot(pen=pg.mkPen(color=(  0,255,255), style=pg.Qt.QtCore.Qt.DashDotLine)),
        }
        for curve in self._profile_y_curves.values():
            curve.rotate(-90)

        # Hide profiles by default
        self._plot_x_profile.widget().hide()
        self._plot_y_profile.widget().hide()
        self._grid.removeItem(self._plot_x_profile)
        self._grid.removeItem(self._plot_y_profile)
        self._grid.removeItem(self._image_widget)
        self._grid.addItem(self._image_widget, 0, 0)


    def set_image_data(self, data, color_mode):
        """
        Make an internal copy of the data image.

        :param data: (Numpy arrray) Image of the data as the Numpy array.
        :param color_mode: (enum) Code of the color mode.
        :return: None
        """
        self._data_mutex.lock()
        self._data = np.copy(data)
        self._data_color_mode = color_mode
        self._data_mutex.unlock()
        self._x_profile, self._y_profile = self._calculate_profiles()


    def plot(self, display_width, display_height):
        """
        Plot the profiles. This method does nothing if profiles are disabled.

        :param display_width: (int) Image width size on display in pixels.
        :param display_height: (int) Image height size on display in pixels.
        :return: None
        """
        # Return if profiles are not shown
        if not self._display_profiles:
            return

        # Calculate profiles
        x_profile, y_profile =  self._x_profile, self._y_profile

        # Setup correct dimensions for the plots
        self._plot_x_profile.widget().setMinimumWidth(display_width)
        self._plot_x_profile.widget().setMaximumWidth(display_width)

        self._plot_y_profile.widget().setMinimumHeight(display_height)
        self._plot_y_profile.widget().setMaximumHeight(display_height)

        self._grid.update()

        # Draw X profile curves
        self._data_mutex.lock()
        if type(x_profile) is list:
            for mode, curve in self._profile_x_curves.items():
                if mode == 'red':
                    curve.setData(x_profile[0])
                    curve.show()
                elif mode == 'green':
                    curve.setData(x_profile[1])
                    curve.show()
                elif mode == 'blue':
                    curve.setData(x_profile[2])
                    curve.show()
                else:
                    curve.hide()
        else:
            for mode, curve in self._profile_x_curves.items():
                if mode == 'mono':
                    curve.setData(x_profile)
                    curve.show()
                else:
                    curve.hide()

        # Draw Y profile curves
        if type(y_profile) is list:
            for mode, curve in self._profile_y_curves.items():
                if mode == 'red':
                    curve.setData(y_profile[0])
                    curve.show()
                elif mode == 'green':
                    curve.setData(y_profile[1])
                    curve.show()
                elif mode == 'blue':
                    curve.setData(y_profile[2])
                    curve.show()
                else:
                    curve.hide()
        else:
            for mode, curve in self._profile_y_curves.items():
                if mode == 'mono':
                    curve.setData(y_profile)
                    curve.show()
                else:
                    curve.hide()
        self._data_mutex.unlock()

    def show(self, flag):
        """
        Hide or show the profiles. This method should be called from the UI thread.

        :param flag: (bool) True to show the profiles, False otherwise.
        :return:
        """
        if flag:
            self._grid.removeItem(self._image_widget)
            self._grid.addItem(self._image_widget, 1, 1)
            self._grid.addItem(self._plot_x_profile, 0, 1, 1, 1, pg.Qt.QtCore.Qt.AlignLeft | pg.Qt.QtCore.Qt.AlignBottom)
            self._grid.addItem(self._plot_y_profile, 1, 0, 1, 1, pg.Qt.QtCore.Qt.AlignRight | pg.Qt.QtCore.Qt.AlignTop)
            self._plot_x_profile.widget().show()
            self._plot_y_profile.widget().show()
            self._display_profiles = True

        else:
            self._grid.removeItem(self._image_widget)
            self._grid.removeItem(self._plot_x_profile)
            self._grid.removeItem(self._plot_y_profile)
            self._grid.addItem(self._image_widget, 1, 1)
            self._plot_x_profile.widget().hide()
            self._plot_y_profile.widget().hide()
            self._display_profiles = False

        self._grid.update()

#########################################
#   Internal methods
#########################################

    def _calculate_profiles(self):
        """
        Calculate the profiles.

        :return: (numpy array, numpy array) X and Y profiles of the image.
        """
        # Calculate the profiles
        self._data_mutex.lock()
        x_profile = y_profile = None
        if self._data_color_mode == COLOR_MODE_MONO:
            x_profile = self._data.mean(axis=1)
            y_profile = self._data.mean(axis=0)
        else:
            x_profile = []
            y_profile = []
            for i in range(3):
                x_profile.append(self._data[:,:,i].mean(axis=1))
                y_profile.append(self._data[:,:,i].mean(axis=0))
        self._data_mutex.unlock()

        return x_profile, y_profile



# -*- coding: utf-8 -*-

"""
Copyright 2018 UChicago Argonne LLC
 as operator of Argonne National Laboratory

PVA object viewer utilities for image display

@author: Guobao Shen <gshen@anl.gov>
"""
from datetime import datetime
import numpy as np
import pyqtgraph
import pyqtgraph.ptime as ptime
from pyqtgraph.Qt import QtCore

class PlotChannel:
    def __init__(self, pvname, color):
        self.color = color
        self.pvname = pvname
        self.dc_offset = 0
        self.axis_location = 'left'
        

class PlotWidget(pyqtgraph.GraphicsWindow):

    def __init__(self, parent=None, **kargs):
        pyqtgraph.GraphicsWindow.__init__(self, parent=parent)
        self.setParent(parent)

        self.param_changed = False
        self.model = None
        self.channels = []
        self.auto_scale = False
        self.first_data = True
        self.plotting_started = False
        
        self.sampling_mode = False        
        self.sample_data = {}
        
        self.mutex = pyqtgraph.QtCore.QMutex()

        self.max_length = None
        self.data_size = 0
        self.new_buffer = True
        self.data = {}
        self.trig_data = {}
        self.first_run = True
        self.new_plot = True
        self.is_max_length = False
        
        # last id number of array received
        self.lastArrayId = None
        # count of lost arrays
        self.lostArrays = 0
        # count of array received
        self.arraysReceived = 0

        # EPICS7 field name for horizontal axis
        self.current_xaxes = "None"
        # EPICS7 field name for Array ID
        self.current_arrayid = "None"

        self.fps = 0
        self.lastTime = ptime.time()

        self._max = None
        self._min = None

        self.bins = 100
        self.fft = False
        self.fft_filter = None
        self.psd = False
        self.diff = False
        self.xy = False
        self.histogram = False
        self.average = 1

        self.__fft_vgain = {
            'none' : 1,
            'hamming' : 1.853,
        }

        self.axis = []
        self.views = []
        self.curves = []

        # Setup plot variables
        self.single_axis = True
        self.plot = None
        self.setup_plot([])


        ##############################
        #
        # trigger related variables
        #
        ##############################
        # Is trigger mode enabled
        self.trigger_mode = False

        # True if in trigger mode and trigger was received (trigger PV monitor fired).
        self.is_triggered = False

        # Counts trigger PV monitor fired callback (count monitor updates)
        self.trigger_count = 0

        # Number samples to take after trigger
        self.samples_after_trig = 0

        # Number of samples captured after trigger
        self.samples_after_trig_cnt = 0

        # Buffer managemet
        self.trigger_buffer_position = 0
        self.samples_before_trig = 0

        self.trigger_data_done = True

        # true of we are monitoring a channel
        self.trigger_is_monitor = False
        # true if we put in bad pv name, and try to monitor
        self.trigger_pv_error = False
        # double sec past epoch timestamp from the trig pv
        self.trigger_timestamp = 0.0

        self.trigger_rec_type = None
        self.trigger_level = 0.0
        ##############################
        #
        # End Trigger mode variables
        #
        ##############################

        self.is_freeze = False

        class TriggerSignal(QtCore.QObject):
            my_signal = QtCore.pyqtSignal()

            def __init__(self):
                QtCore.QObject.__init__(self)

        self.plot_trigger_signal_emitter = TriggerSignal()

    def set_model(self, model):
        """

        :param model:
        :return:
        """
        self.model = model


    def enable_sampling_mode(self, val):
        """
        Set sampling mode.  If true, plot only the latest value at each refresh interval.
        Set this mode if plotting multiple data sources that have different data rates.
        """
        self.sampling_mode = val
        self.data.clear()
        self.setup_plot()
        
    def set_arrayid(self, value):
        """
        Set current field name for array id

        :param value:
        :return:
        """
        if value != self.current_arrayid:
            self.current_arrayid = value

    def set_xaxes(self, value):
        """
        Set current field name for x axes

        :param value:
        :return:
        """
        if value != self.current_xaxes:
            self.current_xaxes = value

    def setup_plot(self, channels=None, single_axis=None):
        """
        Setup plotting

        :param names: list of EPICS7 field names
        :param single_axis: flag to share single axis, or have a axis for each figure. FFT and PSD support only single axis
        :return:
        """
        if single_axis is None:
            single_axis = self.single_axis

        # FFT and PSD support only single axis, PyQTGraph currently doesn't support log scale on the multiple axis setup
        if not single_axis and (self.fft or self.psd):
            single_axis = True
            # Could Be replaced with logging if added in the future
            print(f"{datetime.now().strftime('%Y-%m-%d %H:%M:%S')}: FFT or PSD selected in multi axis mode, which is not supported. Changed to one axis mode automatically.")

        # Delete existing plots
        self.delete_plots()

        # Set new list of signals
        if channels:
            self.channels = channels

        # Create plot item
        self.plot = pyqtgraph.PlotItem()
        self.plot.showGrid(x=True, y=True)
        self.trigMarker = self.plot.plot(pen='r')
        self.do_autoscale()
        if self.fft or self.psd:
            self.plot.setLogMode(x=True, y=True)

        # Generate plot items
        self.single_axis = single_axis
        if single_axis:
            self.setup_plot_single_axis()
        else:
            self.setup_plot_multi_axis()

        # Update plots
        self.plot.vb.sigResized.connect(self.update_views)
        self.update_views()

        # Set for autoscale the first time
        self.new_plot = True
        self.data_size = 0
        
    def setup_plot_single_axis(self):
        """
        Setup items for single Y axis plot.

        :names: (list -> strings) List of channels to bi displayed. "None" should be used to ignore that channel.
        :return: (None)
        """
        if self.channels:
            self.plot.addLegend()
        for ch in self.channels:
            if ch.pvname == "None":
                continue
            curve = self.plot.plot(pen=ch.color, name=ch.pvname)
            curve.plotdata_ave = None
            self.curves.append(curve)

        self.ci.addItem(self.plot, row=2)

    def setup_plot_multi_axis(self):
        """
        Setup items for multiple Y axis plot.

        :names: (list -> strings) List of channels to bi displayed. "None" should be used to ignore that channel.
        :return: (None)
        """
        # We do not use default axis
        self.plot.hideAxis('left')

        # Create axis and views
        left_axis = []
        right_axis = []
        for i, ch in enumerate(self.channels):
            if ch.pvname == "None":
                continue

            axis = pyqtgraph.AxisItem(ch.axis_location)
            axis.setLabel(f"Channel {i+1} [{ch.pvname}]", color=ch.color)
            if ch.axis_location == "left":
                left_axis.append(axis)
            else:
                right_axis.append(axis)
            self.axis.append(axis)
            # First view we take from plotItem
            if len(self.axis) == 1:
                self.views.append(self.plot.vb)
            else:
                self.views.append(pyqtgraph.ViewBox())

        # Add axis and plot to GraphicsLayout (self.ci)
        # Order how they are added is important, while it is also important
        # we hold correct order (the same as in the "names" variable) in self.axis
        # variable as other code depends on this.
        for a in left_axis:
            self.ci.addItem(a, row=2)
        self.ci.addItem(self.plot, row=2)
        for a in right_axis:
            self.ci.addItem(a, row=2)

        # Add Viewboxes to the layout
        for vb in self.views:
            if vb != self.plot.vb:
                self.ci.scene().addItem(vb)

        # Link axis to view boxes
        for i, view in enumerate(self.views):
            self.axis[i].linkToView(view)
            if view != self.plot.vb:
                view.setXLink(self.plot.vb)

        # Make curves
        count = 0

        for i, ch in enumerate(self.channels):
            if ch.pvname == 'None':
                continue
            
            curve = pyqtgraph.PlotCurveItem(pen=ch.color)
            curve.plotdata_ave = None
            left_curve = []
            right_curve = []
            if ch.axis_location == "left":
                left_curve.append(curve)
            else:
                right_curve.append(curve)
            self.curves.extend(left_curve)
            self.curves.extend(right_curve)
            self.views[count].addItem(curve)
            count += 1

    def delete_plots(self):
        """
        Delete all plots

        :return:
        """
        if self.single_axis:
            for curve in self.curves:
                self.plot.removeItem(curve)
        else:
            # Delete old plots if exists
            for i, view in enumerate(self.views):
                view.removeItem(self.curves[i])

            for a in self.axis:
                self.ci.removeItem(a)

        # Remove plotItem
        if self.plot is not None:
            self.ci.removeItem(self.plot)

        # Remove references to all chart items
        self.views = []
        self.axis = []
        self.curves = []

    def update_views(self):
        """
        Update the view so they scale properly.

        :return:
        """
        for view in self.views:
            if view == self.plot.vb:
                continue
            view.setGeometry(self.plot.vb.sceneBoundingRect())

    def set_autoscale(self, flag):
        """
        Enable or disable the autoscale.

        :param flag: (bool) True to enable the autoscale, False otherwise.
        :return:
        """
        self.auto_scale = flag
        self.do_autoscale()


    def do_autoscale(self, auto_scale=None):
        """
        Enable/disable auto range of the current plot based on the `self.auto_scale` setting.

        :param auto_scale: (bool) Can used to overwrite the self.auto_scale. Keep non to use self.auto_scale.
        :return:
        """
        if auto_scale is None:
            auto_scale = self.auto_scale

        if auto_scale:
            self.plot.enableAutoRange()
            for view in self.views:
                view.enableAutoRange()
        else:
            self.plot.disableAutoRange()
            for view in self.views:
                view.disableAutoRange()

    def wait(self):
        """

        :return:
        """
        self.mutex.lock()

    def signal(self):
        """

        :return:
        """
        self.mutex.unlock()

    def update_buffer(self, value):
        """

        :param value:
        :return:
        """
        self.max_length = value
        self.samples_after_trig = int(self.max_length/2)
        self.new_buffer = True
        self.is_max_length = False
        
    def set_range(self, **kwargs):
        """
        Set data range for data filtering.
        Any data out of range will be filtered out, and not plotted.

        :param kwargs:
        :return:
        """
        self._max = kwargs.get("max", None)
        self._min = kwargs.get("min", None)

    def filter(self, data, pair=None):
        """
        Filter data out of range

        :param data:
        :param pair:
        :return:
        """
        rd, rt = data, pair
        if self._max is not None:
            mask = rd <= self._max
            rd = rd[mask]
            if pair is not None:
                rt = rt[mask]
        if self._min is not None:
            mask = rd >= self._min
            rd = rd[mask]
            if pair is not None:
                rt = rt[mask]
        if pair is not None:
            return rd, rt
        return rd

    def set_display_mode(self, value):
        """

        :param value:
        :return:
        """
        if value == 'normal':
            self.fft = False
            self.psd = False
            self.diff = False
            self.xy = False
        elif value == 'fft':
            self.fft = True
            self.psd = False
            self.diff = False
            self.xy = False
        elif value == 'psd':
            self.fft = False
            self.psd = True
            self.diff = False
            self.xy = False
        elif value == 'diff':
            self.fft = False
            self.psd = False
            self.diff = True
            self.xy = False
        elif value == 'xy':
            self.fft = False
            self.psd = False
            self.diff = False
            self.xy = True
        self.plot.autoScale = True

        # Mode changed, iterate over the curves and set moving average to None
        for curve in self.curves:
            curve.plotdata_ave = None

    def set_fft_filter(self, value):
        """
        Set fft filter to use. Valid values are None / "none" and "hamming".

        :param value: (None or string) Filter to be used on the data, before the fft.
        :return:
        """
        if value is None:
            self.fft_filter = value
        else:
            self.fft_filter = value.lower()

    def set_average(self, value):
        """
        Set average number

        :param value:
        :return:
        """
        self.average = value

    def set_histogram(self, flag):
        """
        Set flag to enable/disable histogram plotting

        :param flag:
        :return:
        """
        if flag != self.histogram:
            self.new_plot = True

        self.histogram = flag

    def set_binning(self, value):
        """
        Set number for binning

        :param value:
        :return:
        """
        self.bins = value

    def trigger_process(self, data):
        """
        Process trigger signal. It currently supports EPICS classic record types,
        which are ai/ao, bi/bo, calc, longin/longout, and event.

        It generates a trigger event when it receives an EPICS event and the previous trigger process is done.

        For a bi/bo type trigger, the trigger event is generated at rising edge, which means 0 => 1, and ignores
        the event 1 => 0.

        For a longin/longout type trigger, it always generates a trigger when there is any value change,
        and previous trigger is done.

        For event and calc type trigger, it behaviors the same with longin/longout.

        For a ai/ao type trigger, it generates a trigger event when
            1. its value is greater than give threshold;
            2. previous trigger is done;
            3. triggering at rising edge;
        It will be triggered again once the current plotting is done even for the same trigger edge.
        A special approach shall be adopted to handle the trigger event caused by the same rising edge.

        More algorithm to be added later when needed.
        The current implementation is the first approach, and shall be reconsidered when needed.

        :param data:
        :return:
        """
        # TODO handle trigger event for the case of plotting freezing,
        # need to handle multiple trigger before done with current trigger, and during plotting freezing
        #

        # Count triggers and ignore first one which is initial connection callback and not actual value change.
        self.trigger_count += 1
        if self.trigger_count <= 1:
            return

        # Ignore trigger if we are not ploting
        if not self.plotting_started:
            return

        #  We trigger again only if the last trigger was fully processed
        if not self.trigger_data_done:
            return

        # Process callback
        if self.trigger_rec_type in ["bi", "bo"]:
            # only trigger during jumping from 0 => 1
            if data["value"]["index"] == 1:
                self.is_triggered = True
                self.trigger_data_done = False
        elif self.trigger_rec_type in ["longin", "longout"]:
            # always triggers when values changes
            self.is_triggered = True
            self.trigger_data_done = False
        elif self.trigger_rec_type in ["calc"]:
            # always triggers when values changes
            self.is_triggered = True
            self.trigger_data_done = False
        elif self.trigger_rec_type in ["event"]:
            # always triggers when values changes
            # TODO compare the time stamp to determine the event trigger
            self.is_triggered = True
            self.trigger_data_done = False
        elif self.trigger_rec_type in ["ai", "ao"]:
            # always triggers when values changes
            if data["value"] >= self.trigger_level:
                self.is_triggered = True
                self.trigger_data_done = False
            else:
                self.is_triggered = False

        # TODO: also, if hold off ignore, then during a hold off time, we ignore triggers for example for 1 second.

        if self.is_triggered:
            self.samples_after_trig_cnt = 0
            ts = data['timeStamp']
            self.trigger_timestamp = ts['secondsPastEpoch'] + 1e-9*ts['nanoseconds']

    def __is_trigger_in_array(self, time_array):
        """
        Calculate if the trigger timestap is in the timestamp array.

        :param time_array: (numpy array) Time (SORTED!!!) array to be checked.
        :return: (Tuple -> (bool, integer)) True or False if timestamp is in the array.
                                            If True, second element hold index of the element, otherwise is None.
        """

        idx = np.searchsorted(time_array, self.trigger_timestamp)

        if idx <= 0 or idx >= time_array.size-1:
            return (False, None)
        else:
            return (True, idx)


    def clear_sample_data(self, pvname):
        if pvname in self.sample_data:
            del self.sample_data[pvname]
            if self.sampling_mode and pvname in self.data:
                del self.data[pvname]
                
    def data_process(self, data_generator):
        """
        Process raw data off the wire

        :param data_generator: generator that yields pvaccess data from the wire as key/value pairs
        :return:
        """
        # Increment array count
        self.arraysReceived += 1

        # Check if scope is frozen
        if self.is_freeze:
            return

        self.wait()
        new_size = self.data_size
        vector_len = 0

        # Assume a triggers done then set false when we gather data.
        # all_triggers_done = True

        # We step through all the signals we get from the v4 pv. We keep a count of them
        # start at -1 so we increment to 0 at beginning of loop.
        signal_count = -1

        # The first np array is the one we determine trigger position in the data. So as we step all data vectors
        # with k, the 1st vector we mark as "vector_data_count".
        vector_data_count = -1


        for k, v in data_generator():
            if k == self.current_arrayid:
                if self.lastArrayId is not None:
                    if v - self.lastArrayId > 1:
                        self.lostArrays += 1
                self.lastArrayId = v
                if self.data_size == 0:
                    new_size += 4

            # If pvaccess modules was build without numpy, normal Python list is returned. Make conversion here.
            if type(v) is list:
                v = np.array(v)

            if np.isscalar(v):
                v = np.array([v])
                
            if type(v) is not np.ndarray:
                continue

            # At this point in program we found an nd array that may or may not have data. nd array is
            # not an EPICS7 NDArray, but a numpy ndarray.
            # If no data in a pv field, we just skip this whole loop iteration.
            if len(v) == 0:
                continue

            if not np.isscalar(v[0]):
                continue

            if self.data_size == 0:
                type_size = v[0].dtype.itemsize
                new_size += type_size * len(v)

            if not self.is_max_length and len(self.data.get(k, [])) >= self.max_length:
                self.is_max_length = True

            if self.sampling_mode:
                self.sample_data[k] = v[-1]
            else:
                
                # If we get here, then we found an epics array in the EPICS7 pv that has data.
                # Count array type signals that have data, so triggering works.
                signal_count += 1

                # The 1st time we get here, we actually found k pointing to a data vector,
                # so mark the index of the signal
                if vector_data_count == -1:
                    vector_data_count = signal_count

                # Number of elements in vector
                vector_len = len(v)

                # Here we store the data to class variables for later ploting. How much data we must store depends on the
                # mode we are in.
                if self.trigger_mode:
                    # In trigger mode the minimum we must save is the whole last array (+ part of the old ones if max_length > vector_len)
                    # We must store all the arrayes as the "Time" field determinate how much data do we really need, and the "Time" field can arrive last.
                    required_data = 2 * (vector_len if vector_len > self.max_length else self.max_length)
                    self.data[k] = np.append(self.data.get(k, []), v)[-required_data:]
                else:
                    # We are in free running mode. Only last max_length of the data can be stored.
                    self.data[k] = np.append(self.data.get(k, []), v)[-self.max_length:]

        # Trigger mode is enabled, trigger triggered and we received vector data, execute trigger procedure.
        if (self.trigger_mode
            and self.is_triggered
            and vector_data_count != -1):

            # Check if we have trigger timestamp in buffer
            in_array, idx = self.__is_trigger_in_array(self.data['Time'])
            if in_array:
                self.samples_after_trig_cnt = self.data['Time'].size - idx

                # Check if we have the requested number of samples, to trigger the plotting
                if self.samples_after_trig_cnt >= self.samples_after_trig:
                    for k, v in self.data.items():
                        if type(v) != np.ndarray:
                            continue
                        # `idx` holds the index in the stored waveforms where the trigger happened.
                        # Goal is to get the slice of the waveform so the sample when the trigger happened is in the middle.
                        # samples_after_trig determinate how much samples do we want to show after the trigger (idx + samples_after_trig),
                        # while max_length holds a maximum number of samples to show, for this reason we need `(self.max_length-self.samples_after_trig`
                        # samples before trigger/idx.
                        self.trig_data[k] = v[idx-(self.max_length-self.samples_after_trig) : idx+self.samples_after_trig]
                    self.plot_trigger_signal_emitter.my_signal.emit()

        self.signal()
        if self.data_size == 0:
            self.data_size = new_size
            
        if self.first_data:
            self.first_data = False

    def calculate_ftt(self, data, sample_period, mode, filter=None):
        """
        Calculate fft on the input array.

        :param data: (Numpy array) Data to be transformed.
        :param sample_period: (int) Sampling period of the data.
        :param mode: (string) This could be either "fft" or "psd".
        :param filter: (None or string) Filter to be applied on the data before transformation.

        :return: (Numpy array, Numpy array) xf and yf transformations of the data.
        """
        xf = yf = None
        filter = filter or "none"

        # Perform transformation for X axis
        data_len = len(data)
        xf = np.fft.rfftfreq(data_len, d=sample_period)

        # Calculate window
        if filter == "none":
            pass
        elif filter == "hamming":
            window = np.hamming(data_len)
            data = data * window
        else:
            raise ValueError(f"{str(filter)} is not valid FFT filter type.")

        # Perform transformation for Y axis
        yf_raw = np.abs(np.fft.rfft(data))

        # Apply normalisation filter
        try:
            vertical_gain = self.__fft_vgain[filter]
        except KeyError:
            raise ValueError(f"{str(filter)} is not valid FFT filter type.")

        if mode.lower() == "fft":
            yf = 2. * vertical_gain * yf_raw / data_len
            # DC bin has different scale factor
            yf[0] = yf[0] / 2.
        elif mode.lower() == "psd":
            # Sample period in sec. Sample rate is 1/sample rate. Hz/bin=srage/nf
            sample_rate = 1. / sample_period
            yf = 2. * vertical_gain * vertical_gain * yf_raw * yf_raw / (sample_rate * sample_rate)
            # DC bin has different scale factor
            yf[0] = yf[0] / 4.
        else:
            raise ValueError(f"{str(mode)} is not valid mode for FFT.")

        return xf, yf

    def exponential_moving_average(self, data, data_ave):
        """
        Calculate exponential moving average on the data.

        :param data: (ndarray) Latest array data.
        :param data_ave: (ndarray) Data average until now.

        :return: (ndarray) Exponential moving average.
        """
        alpha = 2. / (self.average + 1.)

        if data_ave is None or len(data_ave) != len(data):
            data_ave=np.array([1e-10] * len(data))

        new_data_ave = data_ave + alpha * (data - data_ave)

        return new_data_ave

    def draw_curve(self, count, data, channel, draw_trig_mark=False):
        """
        Draw a waveform curve

        :param index:  count of curve for plotting
        :param data:   curve data for plotting
        :param draw_trig_mark: flag whether drawing trigger mark
        :return:
        """

        data_len = len(data)
        # in case on time reference in PV, we declare sample period to sec per sample.
        # 1 second per sample as initial
        sample_period = 1.0
        # time array
        time_array = None

        if self.current_xaxes != "None" and len(self.data[self.current_xaxes]) == data_len:
            # TODO later to support: frequency field & sample period as time reference
            # TODO need to handle multiple waveform plotting with different data length
            # Currently, support time only
            sample_period = np.diff(self.data[self.current_xaxes]).mean()
            time_array = self.data[self.current_xaxes]

        if self.diff:
            d = np.diff(data)

        xf = yf = None
        if self.fft or self.psd:
            if data_len == 0:
                return
            mode = "fft" if self.fft else "psd"
            xf, yf = self.calculate_ftt(data, sample_period, mode, self.fft_filter)

        if self.histogram and not self.psd and not self.fft:
            self.curves[count].opts['stepMode'] = True
        else:
            self.curves[count].opts['stepMode'] = False

        # Exponential moving average
        if self.average > 1:
            if yf is not None:
                yf = self.exponential_moving_average(yf, self.curves[count].plotdata_ave)
                self.curves[count].plotdata_ave = yf
            else:
                data = self.exponential_moving_average(data, self.curves[count].plotdata_ave)
                self.curves[count].plotdata_ave = data

        # Exponential moving average
        if self.average > 1:
            if yf is not None:
                yf = self.exponential_moving_average(yf, self.curves[count].plotdata_ave)
                self.curves[count].plotdata_ave = yf
            else:
                data = self.exponential_moving_average(data, self.curves[count].plotdata_ave)
                self.curves[count].plotdata_ave = data

        # Draw curve for the fft or psd mode
        if self.fft or self.psd:
            self.curves[count].setData(xf, yf)

        # Calculate histogram bins and draw the curve
        elif self.histogram and not self.psd and not self.fft:
            d = self.filter(data)
            y, x = np.histogram(d, bins=self.bins)
            self.curves[count].setData(x, y)

        # Draw waveform where X axis are indexes of data array which is drawn on Y axis
        elif time_array is None:
            data_to_plot = self.filter(data) + channel.dc_offset
            self.curves[count].setData(data_to_plot)
            self.__handle_trigger_marker__(draw_trig_mark, self.max_length - self.samples_after_trig, [data_to_plot.min(), data_to_plot.max()])

        # Draw time_array on the X axis and data on Y
        else:
            d, t = self.filter(data, time_array)
            data_to_plot = d + channel.dc_offset
            self.curves[count].setData(t - t[0], data_to_plot)
            self.__handle_trigger_marker__(draw_trig_mark, self.trigger_timestamp - time_array[0], [data_to_plot.min(), data_to_plot.max()])


    def __handle_trigger_marker__(self, draw_trig_mark, marktime, mark_size=None):
        """
        Draw vertical line trigger mark at the specified marktime.

        :param draw_trig_mark: (bool) Flag if trigger line should be drawn.
        :param marktime: (int) Sample number at which the trigger mark happened.
        :param mark_size: (Array [int, int]) Y range for the trigger mark.
        :return:
        """
        if self.trigger_mode and self.is_triggered:
            if draw_trig_mark:
                # Add trigger marker on plotting
                vr = self.plot.viewRange()[1]
                pvnames = [ c.pvname for c in self.channels ] 
                if vr[0] == 0 and vr[1] == 1:
                    min_value = max_value = None
                    if self.single_axis:
                        for field, value in self.trig_data.items():
                            if field in pvnames:
                                a_min = np.amin(value).item()
                                a_max = np.amax(value).item()
                                min_value = a_min if min_value is None else min(min_value, a_min)
                                max_value = a_max if max_value is None else max(max_value, a_max)
                        mark_size = [min_value, max_value]
                    else:
                        mark_size = mark_size
                else:
                     mark_size = vr
                self.trigMarker.setData(np.array([marktime, marktime]),
                                        np.array(mark_size))
        else:
            self.trigMarker.clear()

    def update_drawing(self):
        """
        Update display plotting

        :return:
        """
        if self.first_data or not self.plotting_started:
            # TODO this is to help to over come race condition. To be done in a better way later
            # self.signal()
            return

        # Do not update the drawing if image is frozen
        if self.is_freeze:
            return

        # Take ownership the self.data variable which holds all the waveforms
        self.wait()

        if self.sampling_mode:
            for k, v in self.sample_data.items():
                self.data[k] = np.append(self.data.get(k, []), v)[-self.max_length:]

        # Iterate over the names (channels) selected on the GUI and draw a line for each
        count = 0
        draw_trig_mark = True
        for idx, ch in enumerate(self.channels):
            if ch.pvname != "None":
                if self.trigger_mode:
                    data = self.trig_data.get(ch.pvname)
                else:
                    data = self.data.get(ch.pvname)
                if data is not None:
                    self.draw_curve(count, data, ch, draw_trig_mark)
                    draw_trig_mark = False
                count = count + 1
                

        # Trigger curves were drawn - Freeze the scope
        if self.trigger_mode and self.is_triggered:
            self.is_triggered = False
            self.trigger_data_done = True

        # Axis scaling
        if self.first_run or self.new_buffer or self.new_plot:
            # Perform auto range for the first time
            self.do_autoscale(auto_scale=True)
            self.first_run = False
            self.new_buffer = False
            self.new_plot = False
        elif not self.is_max_length:
            self.do_autoscale(auto_scale=True)
            
        # Then set it to the correct setting
        self.do_autoscale()

        self.update_fps()

        # Release the ownership on the self.data
        self.signal()

    def update_fps(self):
        """
        Update fps statistics

        :return:
        """
        now = ptime.time()
        dt = now - self.lastTime
        self.lastTime = now
        if self.fps is None:
            self.fps = 1.0 / dt
        else:
            s = np.clip(dt * 3., 0, 1)
            self.fps = self.fps * (1 - s) + (1.0 / dt) * s



# -*- coding: utf-8 -*-

from .scope_config_base import ScopeConfigureBase

"""
Copyright 2018 UChicago Argonne LLC
 as operator of Argonne National Laboratory

PVA object viewer utilities

@author: Guobao Shen <gshen@anl.gov>
"""


class Configure(ScopeConfigureBase):
    """
    Scope application configuration panel settings
    """
    def __init__(self, params, **kwargs):
        """

        :param params: parameters parsed from command line and configuration file
        :param pvs: pv name dictionary, format: {"alias": "PV:Name"}
        """
        super().__init__(params, show_start=True, show_trigger=True, **kwargs)
        self.pvs = kwargs.get("pv", None)

        self.counts = 4
        self.default_color = ['#FFFF00', '#FF00FF', '#55FF55', '#00FFFF', '#5555FF',
                              '#5500FF', '#FF5555', '#0000FF', '#FFAA00', '#000000']
        
    def assemble_channel(self, section=None):
        """
        Assemble channel information for plotting

        :param section:
        :return:
        """
        if section is None:
            self.counts = 4
        else:
            # get channel counts to display, 4 by default
            try:
                self.counts = section.getint('COUNT', 4)
                if self.counts > 10:
                    # limit max channel to display
                    self.counts = 10
            except ValueError:
                # TODO: add logging information
                self.counts = 4

        channel = []
        for i in range(self.counts):
            channel.append(
                {"name": "Channel %s" % (i + 1),
                 "type": "group",
                 "children": [
                     {
                         "name": "Color",
                         "type": "color",
                         "value": self.default_color[i],
                         "readonly": True
                     },
                     {"name": "Field", "type": "list", "values": [], "value": "None"},
                     {"name": "DC offset", "type": "float", "value": 0.0},
                     {"name": "Axis location", "type": "list", "values": {
                         "Left" : "left",
                         "Right" : "right",
                     }, "value" : "Left"},
                 ]
                 }
            )

        return channel

    def add_source_acquisition_props(self, children, section):
        pv = None
        if self.pvs is not None:
            pv = list(self.pvs.values())[0]

        if section:
            if pv is None:
                # it means PV map is not specified from command line
                # get one from configuration
                pv = pv if pv is not None else section.get("PV", None)
                if pv is not None:
                    # if PV is available by default
                    self.pvs = {pv: pv}

        newchildren = [
            # EPICS7 PV name, which assumes pvAccess protocol
            # Alias name to be supported later
            {"name": "PV", "type": "str", "value": pv},
            {"name": "PV status", "type": "str", "value": "Disconnected", "readonly": True}
        ]

        newchildren += children
        return newchildren

    def parse(self):
        """

        :return:
        """
        try:
            sections = self.params["SCOPE"]["SECTION"].split(",")
            for idx, section in enumerate(sections):
                # remove unnecessary space
                sections[idx] = section.strip()
            if "ACQUISITION" in sections:
                acquisition = self.assemble_acquisition(self.params["ACQUISITION"])
            else:
                acquisition = self.assemble_acquisition()
            if "DISPLAY" in sections:
                display = self.assemble_display(self.params["DISPLAY"])
            else:
                display = self.assemble_display()
            if "CHANNELS" in sections:
                channel = self.assemble_channel(self.params["CHANNELS"])
            else:
                channel = self.assemble_channel()
        except KeyError:
            acquisition = self.assemble_acquisition()
            display = self.assemble_display()
            channel = self.assemble_channel()

        if acquisition is None or display is None or channel is None:
            raise RuntimeError("No enough information for scope")

        cfg = self.assemble_config()
        # line up in order
        paramcfg = [acquisition, display, cfg]
        for ch in channel:
            paramcfg.append(ch)
        statistics = self.assemble_statistics()
        
        paramcfg.append(statistics)

        return paramcfg



import sys
import os
import psutil
from pyqtgraph.Qt import QtWidgets
from pyqtgraph.Qt import uic
from pyqtgraph.parametertree import Parameter
from .model import DataSource as DataReceiver
from .control import StripToolController
from .view import StripToolConfigure

form_path = os.path.join(os.path.dirname(__file__), "ui/striptool.ui")
form_class = uic.loadUiType(form_path)[0]

class StripToolWindow(QtWidgets.QMainWindow, form_class):
    def __init__(self, parent=None):
        super(StripToolWindow, self).__init__(parent=parent)
        self._proc = psutil.Process()
        self.setupUi(self)
        self.show()

pvedit_dialog_path = os.path.join(os.path.dirname(__file__), "ui/addpv.ui")
pvedit_dialog_class = uic.loadUiType(pvedit_dialog_path)[0]

class PvEditDialog(QtWidgets.QDialog, pvedit_dialog_class):
    def __init__(self, parent=None):
        super(PvEditDialog, self).__init__(parent=parent)
        self.setupUi(self)

warning_path = os.path.join(os.path.dirname(__file__), "ui/warning.ui")
warning_class = uic.loadUiType(warning_path)[0]

class WarningDialog(QtWidgets.QDialog, warning_class):
    def __init__(self, parent=None):
        super(WarningDialog, self).__init__(parent=parent)
        self.setupUi(self)

def striptool(cfg, **kwargs):
    app = QtWidgets.QApplication.instance()
    if app is None:
        app = QtWidgets.QApplication(sys.argv)
    else:
        print('QApplication instance already exists: %s' % str(app))

    w = StripToolWindow()
    model = DataReceiver()
    warning = WarningDialog()
    pvedit_dialog = PvEditDialog()

    configure = StripToolConfigure(cfg, **kwargs)
    parameters = Parameter.create(
        name="params", type="group", children=configure.parse())
    w.parameterPane.setParameters(parameters, showTop=False)

    controller = StripToolController(w, model, pvedit_dialog, warning, parameters, cfg, **kwargs)
    parameters.sigTreeStateChanged.connect(controller.parameter_change)
    
    w.show()
    sys.exit(app.exec_())



import unittest
from configparser import ConfigParser
from c2dataviewer.control.striptool_config import StriptoolConfig
from c2dataviewer.control.pvconfig import PvConfig


class TestStriptoolConfig(unittest.TestCase):
    def test_config(self):
        raw = """
[STRIPTOOL]
DefaultProtocol = ca
Chan1.PV = Foo:Bar
Chan2.PV = pva://Bar:Baz
Chan1.Color = #000000
Chan2.Color = #0000FF
"""
        parser = ConfigParser()
        parser.read_string(raw)
        cfg = StriptoolConfig(parser)
        expected = {
            'Foo:Bar': PvConfig('Foo:Bar', '#000000', 'ca'),
            'Bar:Baz': PvConfig('Bar:Baz', '#0000FF', 'pva')
        }

        self.assertEqual(len(cfg.pvs), len(expected))

        for pv in cfg.pvs.values():
            self.assertTrue(pv.pvname in expected)
            e = expected[pv.pvname]
            self.assertEqual(e.pvname, pv.pvname)
            self.assertEqual(e.color, pv.color)
            self.assertEqual(e.proto, pv.proto)
            del expected[pv.pvname]



import numpy as np
import pyqtgraph
import pvaccess as pva
from ..model import ConnectionState

class ScopeControllerBase:
    def __init__(self, widget, model, parameters, warning, channels=[], **kwargs):
        self._win = widget
        self.model = model
        self.parameters = parameters
        self.data = None
        # refresh frequency: every 100 ms by default
        self.refresh = 100

        self.plotting_started = False

        self.timer = pyqtgraph.QtCore.QTimer()
        self.timer.timeout.connect(self._win.graphicsWidget.update_drawing)
        self._win.graphicsWidget.set_model(self.model)

        self._warning = warning
        self._warning.warningConfirmButton.clicked.connect(lambda: self.accept_warning())

        self.arrays = np.array([])
        self.lastArrays = 0

        self.default_arrayid = "None"
        self.default_xaxes = "None"
        self.current_arrayid = "None"
        self.current_xaxes = "None"
        self.default_trigger = None

        self._win.graphicsWidget.setup_plot(channels=channels, single_axis=True)
        
        # timer to update status with statistics data
        self.status_timer = pyqtgraph.QtCore.QTimer()
        self.status_timer.timeout.connect(self.update_status)
        self.status_timer.start(1000)

    def default_config(self, **kwargs):
        """
        Default configuration for array ID and x axes field names

        :param kwargs:
        :return:
        """
        self.default_arrayid = kwargs.get("arrayid", None)
        if self.default_arrayid is None:
            self.default_arrayid = "None"
        else:
            self._win.graphicsWidget.set_arrayid(self.default_arrayid)
        self.default_xaxes = kwargs.get("xaxes", None)
        if self.default_xaxes is None:
            self.default_xaxes = "None"
        else:
            self._win.graphicsWidget.set_xaxes(self.default_xaxes)
        self._win.graphicsWidget.set_range(**kwargs)

        max_length = self.parameters.child("Acquisition").child("Buffer (Samples)").value()
        if max_length:
            self._win.graphicsWidget.update_buffer(int(max_length))
            
        self._win.graphicsWidget.set_binning(self.parameters.child("Display").child("Num Bins").value())
        
        self.default_trigger = kwargs.get("trigger", None)
        if self.default_trigger is not None:
            try:
                if "://" in self.default_trigger:
                    # pv comes with format of proto://pvname
                    p, name = self.default_trigger.split("://")
                    trt = self.model.update_trigger(name, proto=p.lower())
                else:
                    # PV name only, use default Channel Access protocol
                    trt = self.model.update_trigger(self.default_trigger)
                self.parameters.child("Acquisition").child('Trigger Mode').setWritable()
                self._win.graphicsWidget.trigger_rec_type = trt
                self.__trigger_level_rw__(trt)
            except Exception as e:
                self._win.graphicsWidget.trigger_rec_type = None
                self.parameters.child("Acquisition").child('Trigger PV').setValue("")
                self.parameters.child("Acquisition").child('Trigger Mode').setReadonly()
                self.default_trigger = None
                raise e

            
    def update_buffer(self, size):
        self._win.graphicsWidget.update_buffer(size)
        self.parameters.child("Acquisition").child("Buffer (Samples)").setValue(self._win.graphicsWidget.max_length)


    def __trigger_level_rw__(self, trt):
        """
        Set trigger level parameter readable/writable according trigger record type.
            
        Making trigger level writable is not supported in pyqtgraph 0.10. See
        https://github.com/pyqtgraph/pyqtgraph/issues/263
        This was fixed with 0.11, but we had other issues with the new version. See
        https://git.aps.anl.gov/C2/conda/data-viewer/-/merge_requests/30#note_4799
            
        :param trt: (string) Record type as string (e.g. ai, bi, calc, ...).
        :return:
        """
        pass
    # if trt in ["ai", "ao"]:
    #     self.parameters.child("Acquisition").child('Trigger Threshold').setWritable()
    # else:
    #     self.parameters.child("Acquisition").child('Trigger Threshold').setReadonly()

    def parameter_change(self, params, changes):
        """
        
        :param params:
        :param changes:
        :return:
        """
        for param, change, data in changes:
            if change == "value":
                path = self.parameters.childPath(param)
                if path is not None:
                    childName = '.'.join(path)
                else:
                    childName = param.name()

                if childName == "Acquisition.Trigger PV":
                    if data is None:
                        return
                    if self._win.graphicsWidget.plotting_started:
                        self.notify_warning("Please stop plotting first before changing trigger PV")
                        return
                    data = data.strip()
                    if data != "":
                        try:
                            if "://" in data:
                                # pv comes with format of proto://pvname
                                p, name = data.split("://")
                                trt = self.model.update_trigger(name, proto=p.lower())
                            else:
                                # PV name only, use default Channel Access protocol
                                trt = self.model.update_trigger(data)
                                self.parameters.child("Acquisition").child('Trigger Mode').setWritable()
                                self._win.graphicsWidget.trigger_rec_type = trt
                                self.__trigger_level_rw__(trt)

                            # restart trigger if trigger is enabled
                            if self._win.graphicsWidget.trigger_mode:
                                self.stop_trigger_mode()
                                self.start_trigger_mode()
                        except RuntimeError as e:
                            self._win.graphicsWidget.trigger_rec_type = None
                            self.notify_warning(repr(e))
                            self.parameters.child("Acquisition").child('Trigger Mode').setReadonly()
                            self.stop_trigger_mode()
                            # TODO clear trigger PV text field
                        except Exception as e:
                            self._win.graphicsWidget.trigger_rec_type = None
                            self.notify_warning("Channel {} timed out. \n{}".format(data, repr(e)))
                            self.stop_trigger_mode()
                            # TODO clear trigger PV text field
                elif childName == "Acquisition.Trigger Mode":
                    self.set_trigger_mode(data)
                elif childName == "Acquisition.Trigger Threshold":
                    self._win.graphicsWidget.trigger_level = data
                elif childName == "Acquisition.PostTrigger":
                    self.set_post_tigger(data)
                elif childName == "Acquisition.HoldTrigger":
                    self.set_hold_trigger(data)
                elif childName == "Acquisition.Buffer (Samples)":
                    self._win.graphicsWidget.update_buffer(data)
                elif childName == "Acquisition.Freeze":
                    self._win.graphicsWidget.is_freeze = data
                elif childName == "Display.Mode":
                    self._win.graphicsWidget.set_display_mode(data)
                    self._win.graphicsWidget.setup_plot()
                    # Disable multiaxis for the FFT and PSD modes. As at the time of the
                    # writing this code pyqtgraph does not support logarithmic scale in multiaxis configuration.
                    if self._win.graphicsWidget.fft or self._win.graphicsWidget.psd:
                        self.parameters.child("Display").child("Single axis").setReadonly()
                    else:
                        self.parameters.child("Display").child("Single axis").setWritable()
                elif childName == "Display.FFT filter":
                    self._win.graphicsWidget.set_fft_filter(data)
                elif childName == "Display.Exp moving avg":
                    self._win.graphicsWidget.set_average(data)
                elif childName == "Display.Autoscale":
                    self._win.graphicsWidget.set_autoscale(data)
                elif childName == "Display.Single axis":
                    self._win.graphicsWidget.setup_plot(single_axis=data)
                elif childName == "Display.Histogram":
                    self._win.graphicsWidget.set_histogram(data)
                elif childName == "Display.Num Bins":
                    self._win.graphicsWidget.set_binning(data)
                elif childName == "Display.Refresh":
                    self.set_freshrate(data)
                elif childName == "Config.ArrayId":
                    self.set_arrayid(data)
                elif childName == "Config.X Axes":
                    self.set_xaxes(data)

    def set_freshrate(self, value):
        """
        Set time to refresh
            
        :param value: time interval to plot, in second
        :return:
        """
        plotting_started = self._win.graphicsWidget.plotting_started
        self.stop_plotting()
        self.refresh = value*1000.0
        if plotting_started:
            self.start_plotting()


    def start_plotting(self):
        """
            
        :return:
        """
        try:
            self.timer.timeout.disconnect()
            self.timer.stop()
        except Exception:
            pass

        # Disable trigger signal
        if self._win.graphicsWidget.plot_trigger_signal_emitter is not None:
            try:
                self._win.graphicsWidget.plot_trigger_signal_emitter.my_signal.disconnect()
            except Exception:
                pass

        # Setup free run plotting
        if not self._win.graphicsWidget.trigger_mode:
            self.timer.timeout.connect(self._win.graphicsWidget.update_drawing)
            self.timer.start(self.refresh)

        # Setup trigger plotting
        if self._win.graphicsWidget.trigger_mode and self._win.graphicsWidget.plot_trigger_signal_emitter is not None:
            self._win.graphicsWidget.plot_trigger_signal_emitter.my_signal.connect(self._win.graphicsWidget.update_drawing)

        self._win.graphicsWidget.plotting_started = True
                
    def stop_plotting(self):
        """

        :return:
        """
        self._win.graphicsWidget.plotting_started = False

        # Stop timer
        self.timer.stop()

        # Stop signal emitter for trigger mode
        if self._win.graphicsWidget.plot_trigger_signal_emitter is not None:
            try:
                self._win.graphicsWidget.plot_trigger_signal_emitter.my_signal.disconnect()
            except Exception:
                pass

            
    def set_arrayid(self, value):
        """
        Set current field name for array id
            
        :param value:
        :return:
        """
        if value != self.current_arrayid:
            self.current_arrayid = value
            self._win.graphicsWidget.current_arrayid = value

    def set_xaxes(self, value):
        """
        Set current field name for x axes
            
        :param value:
        :return:
        """
        if value != self.current_xaxes:
            self.current_xaxes = value
            self._win.graphicsWidget.current_xaxes = value

    def set_trigger_mode(self, value):
        """
        Set trigger mode.
    
        :param value:
        :return:
        """
        self.stop_plotting()
        self._win.graphicsWidget.trigger_mode = value
        if not self._win.graphicsWidget.trigger_mode:
            self._win.graphicsWidget.trigger_data_done = True

        if self.model.trigger_chan is not None:
            if self._win.graphicsWidget.trigger_mode:
                self.start_trigger_mode()
            else:
                self.stop_trigger_mode()
        else:
            self.parameters.child("Acquisition").child('Trigger Mode').setReadonly()

        if self._win.graphicsWidget.plotting_started:
            self.start_plotting()

    def start_trigger_mode(self):
        """
        Process to start DAQ in trigger mode
            
        :return:
        """
        self._win.graphicsWidget.samples_after_trig_cnt = 0

        if not self._win.graphicsWidget.trigger_is_monitor:
            try:
                self.model.start_trigger(self._win.graphicsWidget.trigger_process)
                self._win.graphicsWidget.trigger_count = 0
                self._win.graphicsWidget.trigger_is_monitor = True
            except Exception as e:
                self._win.graphicsWidget.trigger_is_monitor = False
                raise e

        self._win.graphicsWidget.is_triggered = False

    def stop_trigger_mode(self):
        """
        Stop trigger mode
        :return:
        """
        self._win.graphicsWidget.trigger_is_monitor = False
        self._win.graphicsWidget.trigger_count = 0
        if self.model.trigger_chan is not None:
            self.model.stop_trigger()

    def set_post_trigger(self, value):
        """
        Set the time period after the trigger.
        
        :param value:
        :return:
        """
        # TODO need to understand this requirement more to implement it
        # it is currently a place holder

    def set_hold_trigger(self, value):
        """
        Set time period to hold the trigger
            
        :param value:
        :return:
        """
        # TODO need to understand this requirement more to implement it
        # it is currently a place holder

    def accept_warning(self):
        """
            
        :return:
        """
        self._warning.close()

    def notify_warning(self, msg):
        #close previous warning
        self.accept_warning()
            
        self._warning.warningTextBrowse.setText(msg)
        self._warning.show()

        
    def update_status(self):
        """
        Update statistics status.
            
        :return:
        """
        # Update display
        single_axis_child = self.parameters.child("Display", "Single axis")
        if single_axis_child.value() != self._win.graphicsWidget.single_axis:
            single_axis_child.setValue(self._win.graphicsWidget.single_axis)

        # Update statistics
        with self._win._proc.oneshot():
            cpu = self._win._proc.cpu_percent(None)

        # TODO algorithm to calculate Array/sec
        arraysReceived = self._win.graphicsWidget.arraysReceived
        n = arraysReceived - self.lastArrays
        self.lastArrays = arraysReceived
        self.arrays = np.append(self.arrays, n)[-10:]

        for q in self.parameters.child("Statistics").children():
            if q.name() == 'CPU':
                q.setValue(cpu)
            elif q.name() == 'Lost Arrays':
                q.setValue(self._win.graphicsWidget.lostArrays)
            elif q.name() == 'Tot. Arrays':
                q.setValue(self._win.graphicsWidget.arraysReceived)
            elif q.name() == 'Arrays/Sec':
                q.setValue(self.arrays.mean())
            elif q.name() == 'Bytes/Sec':
                q.setValue(self.arrays.mean() * self._win.graphicsWidget.data_size)
            elif q.name() == 'Rate':
                q.setValue(self._win.graphicsWidget.fps)
            elif q.name() == 'TrigStatus':
                stat_str = "Not Trig Mode,Not Monitoring"
                self._win.graphicsWidget.monitoring_trigger = False
                if self._win.graphicsWidget.monitoring_trigger:
                    stat_str = "Not Trig Mode, Monitoring Trigger PV"
                    if self._win.graphicsWidget.trigger_mode:
                        stat_str = "Waiting for Trigger, Collecting"
                        if self._win.graphicsWidget.is_triggered:
                            stat_str = "Got Trigger, Collecting"
                            if self._win.graphicsWidget.trigger_data_done:
                                stat_str = "Got Trigger, Done Collecting"
                q.setValue(stat_str)



# -*- coding: utf-8 -*-

"""
Copyright 2020 UChicago Argonne LLC
 as operator of Argonne National Laboratory

Unit tests for Scope application (scope_display.py)

@author: Matic Pogacnik <mpogacnik@anl.gov>
"""
import os
import sys
import unittest
import numpy as np

from c2dataviewer.view.scope_display import PlotWidget
from c2dataviewer.view.scope_display import PlotChannel

os.environ["QT_QPA_PLATFORM"] = "offscreen"

array1 = [
0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9,
10.0, 11.1, 12.2, 13.3, 14.4, 15.5, 16.6, 17.7, 18.8, 19.9,
20.0, 21.1, 22.2, 23.3, 24.4, 25.5, 26.6, 27.7, 28.8, 29.9,
30.0, 31.1, 32.2, 33.3, 34.4, 35.5, 36.6, 37.7, 38.8, 39.9,
40.0, 41.1, 42.2, 43.3, 44.4, 45.5, 46.6, 47.7, 48.8, 49.9,
]

class TestScopeDisplay(unittest.TestCase):

    def setUp(self):
        """
        Build the environment for each unit test case.
        This method is called before each test.

        :return:
        """
        # Create instance of the PlotWidget
        self.pw = PlotWidget()
        self.pw.update_buffer(100)
        
    def tearDown(self):
        """
        Tear down the environment after each test case.
        This mentod is called after each test.

        :return:
        """
        pass


    def test_no_filter(self):
        """
        Calculate fft without any filter

        :return:
        """

        # Prepare data
        data = np.array(array1, dtype=np.float64)
        sample_period = np.diff(data).mean()

        # Calculate fft
        xf, yf = self.pw.calculate_ftt(data, sample_period, "fft", None)

        # Test result
        np.testing.assert_array_almost_equal(xf,
                                             np.array([0.,         0.01963928, 0.03927856, 0.05891784, 0.07855711, 0.09819639,
                                                       0.11783567, 0.13747495, 0.15711423, 0.17675351, 0.19639279, 0.21603206,
                                                       0.23567134, 0.25531062, 0.2749499 , 0.29458918, 0.31422846, 0.33386774,
                                                       0.35350701, 0.37314629, 0.39278557, 0.41242485, 0.43206413, 0.45170341,
                                                       0.47134269, 0.49098196,], dtype=np.float64),
                                            decimal=6)

        np.testing.assert_array_almost_equal(yf,
                                            np.array([24.95 , 15.926,  7.979,  5.337,  4.021,  3.56 ,  2.716,  2.349,
                                                       2.076,  1.866,  1.871,  1.569,  1.461,  1.372,  1.298,  1.36 ,
                                                       1.184,  1.141,  1.105,  1.076,  1.157,  1.032,  1.018,  1.008,
                                                       1.002,  1.1  ],
                                                    dtype=np.float64),
                                            decimal=3
                                        )

    def test_hamming_filter(self):
        """
        Calculate fft with Hamming filter

        :return:
        """

        # Prepare data
        data = np.array(array1, dtype=np.float64)
        sample_period = np.diff(data).mean()

        # Run transformation
        xf, yf = self.pw.calculate_ftt(data, sample_period, "psd", "hamming")

        # Check results
        np.testing.assert_array_almost_equal(xf,
                                             np.array([0.,         0.01963928, 0.03927856, 0.05891784, 0.07855711, 0.09819639,
                                                       0.11783567, 0.13747495, 0.15711423, 0.17675351, 0.19639279, 0.21603206,
                                                       0.23567134, 0.25531062, 0.2749499 , 0.29458918, 0.31422846, 0.33386774,
                                                       0.35350701, 0.37314629, 0.39278557, 0.41242485, 0.43206413, 0.45170341,
                                                       0.47134269, 0.49098196,], dtype=np.float64),
                                            decimal=6)
        np.testing.assert_array_almost_equal(yf,
                                            np.array([7.80679385e+05, 7.88967744e+05, 2.35308112e+03, 6.03251068e+01,
                                                      6.61654009e+01, 6.03079690e+02, 5.17923210e+01, 1.25620540e+02,
                                                      1.04688807e+02, 4.53016738e+01, 2.15743462e+02, 2.96731584e+01,
                                                      5.81910126e+01, 5.19148396e+01, 2.40652017e+01, 1.18719727e+02,
                                                      1.89395969e+01, 3.70000838e+01, 3.48239178e+01, 1.67868855e+01,
                                                      8.69944055e+01, 1.50511865e+01, 2.98263438e+01, 2.92627440e+01,
                                                      1.44786076e+01, 7.89336302e+01,],
                                                      dtype=np.float64),
                                            decimal=3
                                        )

    def test_exp_moving_average(self):
        """
        Calculate Exponential moving average.

        :return:
        """

        # Prepare data
        data = [ np.array(array1, dtype=np.float64) * d for d in range(1, 7) ]
        average = None

        self.pw.set_average(4)

        # Calculate moving averages
        average = self.pw.exponential_moving_average(data[0], average)
        np.testing.assert_array_almost_equal(average,
                                            np.array(
                                                [6.000e-11, 4.400e-01, 8.800e-01, 1.320e+00, 1.760e+00, 2.200e+00,
                                                2.640e+00, 3.080e+00, 3.520e+00, 3.960e+00, 4.000e+00, 4.440e+00,
                                                4.880e+00, 5.320e+00, 5.760e+00, 6.200e+00, 6.640e+00, 7.080e+00,
                                                7.520e+00, 7.960e+00, 8.000e+00, 8.440e+00, 8.880e+00, 9.320e+00,
                                                9.760e+00, 1.020e+01, 1.064e+01, 1.108e+01, 1.152e+01, 1.196e+01,
                                                1.200e+01, 1.244e+01, 1.288e+01, 1.332e+01, 1.376e+01, 1.420e+01,
                                                1.464e+01, 1.508e+01, 1.552e+01, 1.596e+01, 1.600e+01, 1.644e+01,
                                                1.688e+01, 1.732e+01, 1.776e+01, 1.820e+01, 1.864e+01, 1.908e+01,
                                                1.952e+01, 1.996e+01],
                                              dtype=np.float64),
                                              decimal=3)

        average = self.pw.exponential_moving_average(data[1], average)
        np.testing.assert_array_almost_equal(average,
                                            np.array(
                                                [3.6000e-11, 1.1440e+00, 2.2880e+00, 3.4320e+00, 4.5760e+00,
                                                5.7200e+00, 6.8640e+00, 8.0080e+00, 9.1520e+00, 1.0296e+01,
                                                1.0400e+01, 1.1544e+01, 1.2688e+01, 1.3832e+01, 1.4976e+01,
                                                1.6120e+01, 1.7264e+01, 1.8408e+01, 1.9552e+01, 2.0696e+01,
                                                2.0800e+01, 2.1944e+01, 2.3088e+01, 2.4232e+01, 2.5376e+01,
                                                2.6520e+01, 2.7664e+01, 2.8808e+01, 2.9952e+01, 3.1096e+01,
                                                3.1200e+01, 3.2344e+01, 3.3488e+01, 3.4632e+01, 3.5776e+01,
                                                3.6920e+01, 3.8064e+01, 3.9208e+01, 4.0352e+01, 4.1496e+01,
                                                4.1600e+01, 4.2744e+01, 4.3888e+01, 4.5032e+01, 4.6176e+01,
                                                4.7320e+01, 4.8464e+01, 4.9608e+01, 5.0752e+01, 5.1896e+01],
                                              dtype=np.float64),
                                              decimal=3)

        average = self.pw.exponential_moving_average(data[2], average)
        np.testing.assert_array_almost_equal(average,
                                            np.array(
                                                [2.16000e-11, 2.00640e+00, 4.01280e+00, 6.01920e+00, 8.02560e+00,
                                                1.00320e+01, 1.20384e+01, 1.40448e+01, 1.60512e+01, 1.80576e+01,
                                                1.82400e+01, 2.02464e+01, 2.22528e+01, 2.42592e+01, 2.62656e+01,
                                                2.82720e+01, 3.02784e+01, 3.22848e+01, 3.42912e+01, 3.62976e+01,
                                                3.64800e+01, 3.84864e+01, 4.04928e+01, 4.24992e+01, 4.45056e+01,
                                                4.65120e+01, 4.85184e+01, 5.05248e+01, 5.25312e+01, 5.45376e+01,
                                                5.47200e+01, 5.67264e+01, 5.87328e+01, 6.07392e+01, 6.27456e+01,
                                                6.47520e+01, 6.67584e+01, 6.87648e+01, 7.07712e+01, 7.27776e+01,
                                                7.29600e+01, 7.49664e+01, 7.69728e+01, 7.89792e+01, 8.09856e+01,
                                                8.29920e+01, 8.49984e+01, 8.70048e+01, 8.90112e+01, 9.10176e+01],
                                              dtype=np.float64),
                                              decimal=3)

        average = self.pw.exponential_moving_average(data[3], average)
        np.testing.assert_array_almost_equal(average,
                                            np.array(
                                                [1.2960000e-11, 2.9638400e+00, 5.9276800e+00, 8.8915200e+00,
                                                1.1855360e+01, 1.4819200e+01, 1.7783040e+01, 2.0746880e+01,
                                                2.3710720e+01, 2.6674560e+01, 2.6944000e+01, 2.9907840e+01,
                                                3.2871680e+01, 3.5835520e+01, 3.8799360e+01, 4.1763200e+01,
                                                4.4727040e+01, 4.7690880e+01, 5.0654720e+01, 5.3618560e+01,
                                                5.3888000e+01, 5.6851840e+01, 5.9815680e+01, 6.2779520e+01,
                                                6.5743360e+01, 6.8707200e+01, 7.1671040e+01, 7.4634880e+01,
                                                7.7598720e+01, 8.0562560e+01, 8.0832000e+01, 8.3795840e+01,
                                                8.6759680e+01, 8.9723520e+01, 9.2687360e+01, 9.5651200e+01,
                                                9.8615040e+01, 1.0157888e+02, 1.0454272e+02, 1.0750656e+02,
                                                1.0777600e+02, 1.1073984e+02, 1.1370368e+02, 1.1666752e+02,
                                                1.1963136e+02, 1.2259520e+02, 1.2555904e+02, 1.2852288e+02,
                                                1.3148672e+02, 1.3445056e+02],
                                              dtype=np.float64),
                                              decimal=3)

        average = self.pw.exponential_moving_average(data[4], average)
        np.testing.assert_array_almost_equal(average,
                                            np.array(
                                                [7.77600000e-12, 3.97830400e+00, 7.95660800e+00, 1.19349120e+01,
                                                1.59132160e+01, 1.98915200e+01, 2.38698240e+01, 2.78481280e+01,
                                                3.18264320e+01, 3.58047360e+01, 3.61664000e+01, 4.01447040e+01,
                                                4.41230080e+01, 4.81013120e+01, 5.20796160e+01, 5.60579200e+01,
                                                6.00362240e+01, 6.40145280e+01, 6.79928320e+01, 7.19711360e+01,
                                                7.23328000e+01, 7.63111040e+01, 8.02894080e+01, 8.42677120e+01,
                                                8.82460160e+01, 9.22243200e+01, 9.62026240e+01, 1.00180928e+02,
                                                1.04159232e+02, 1.08137536e+02, 1.08499200e+02, 1.12477504e+02,
                                                1.16455808e+02, 1.20434112e+02, 1.24412416e+02, 1.28390720e+02,
                                                1.32369024e+02, 1.36347328e+02, 1.40325632e+02, 1.44303936e+02,
                                                1.44665600e+02, 1.48643904e+02, 1.52622208e+02, 1.56600512e+02,
                                                1.60578816e+02, 1.64557120e+02, 1.68535424e+02, 1.72513728e+02,
                                                1.76492032e+02, 1.80470336e+02],
                                              dtype=np.float64),
                                              decimal=3)

        average = self.pw.exponential_moving_average(data[5], average)
        np.testing.assert_array_almost_equal(average,
                                            np.array(
                                                [4.66560000e-12, 5.02698240e+00, 1.00539648e+01, 1.50809472e+01,
                                                2.01079296e+01, 2.51349120e+01, 3.01618944e+01, 3.51888768e+01,
                                                4.02158592e+01, 4.52428416e+01, 4.56998400e+01, 5.07268224e+01,
                                                5.57538048e+01, 6.07807872e+01, 6.58077696e+01, 7.08347520e+01,
                                                7.58617344e+01, 8.08887168e+01, 8.59156992e+01, 9.09426816e+01,
                                                9.13996800e+01, 9.64266624e+01, 1.01453645e+02, 1.06480627e+02,
                                                1.11507610e+02, 1.16534592e+02, 1.21561574e+02, 1.26588557e+02,
                                                1.31615539e+02, 1.36642522e+02, 1.37099520e+02, 1.42126502e+02,
                                                1.47153485e+02, 1.52180467e+02, 1.57207450e+02, 1.62234432e+02,
                                                1.67261414e+02, 1.72288397e+02, 1.77315379e+02, 1.82342362e+02,
                                                1.82799360e+02, 1.87826342e+02, 1.92853325e+02, 1.97880307e+02,
                                                2.02907290e+02, 2.07934272e+02, 2.12961254e+02, 2.17988237e+02,
                                                2.23015219e+02, 2.28042202e+02],
                                              dtype=np.float64),
                                              decimal=3)

    def test_multi_axis_plot(self):
        """
        Test multi axis plot.

        :return:
        """
        signals = [PlotChannel('x1', '#0000FF')] * 4
        self.pw.setup_plot(signals, single_axis=False)

        self.assertIsNotNone(self.pw.plot)
        self.assertEqual(len(signals), len(self.pw.axis))
        self.assertEqual(len(signals), len(self.pw.views))
        self.assertEqual(len(signals), len(self.pw.curves))

        
    def test_sampling_mode(self):
        self.pw.enable_sampling_mode(True)
        self.pw.setup_plot([PlotChannel('ch1', '#000000')])
        self.pw.plotting_started = True
        val = 0
        def generator():
            yield 'ch1', val
            
        self.pw.data_process(generator)
        val = 1
        self.pw.data_process(generator)
        self.pw.update_drawing()
        val = 2
        self.pw.data_process(generator)
        val = 3
        self.pw.data_process(generator)
        self.pw.update_drawing()
        self.pw.update_drawing()

        expected = np.array([1,3,3])
        actual = self.pw.data['ch1']
        self.assertTrue((actual == expected).all())
        
        
    def test_trigger_process(self):
        """
        Test "trigger_process" method which is used as the callback for the trigger `camonitor` event.

        :return:
        """

        # Test bi record
        self.pw.trigger_rec_type = "bi"
        data = {}
        data['value'] = {}
        data['value']['index'] = 1
        data['timeStamp'] = {}
        data['timeStamp']['secondsPastEpoch'] = 111
        data['timeStamp']['nanoseconds'] = 222
        ts = data['timeStamp']['secondsPastEpoch'] + 1e-9*data['timeStamp']['nanoseconds']

        self.pw.plotting_started = True

        # Test init state
        self.assertTrue(self.pw.plotting_started)
        self.assertFalse(self.pw.is_triggered)
        self.assertTrue(self.pw.trigger_data_done)
        self.assertAlmostEqual(0.0, self.pw.trigger_timestamp)

        # Run and test state after first, trigger as the first monitor should be
        # ignored because of the init connection call
        self.pw.trigger_process(data)
        self.assertFalse(self.pw.is_triggered)
        self.assertTrue(self.pw.trigger_data_done)
        self.assertAlmostEqual(0.0, self.pw.trigger_timestamp)

        # Call the callback again and check if values were set correctly
        self.pw.trigger_process(data)
        self.assertTrue(self.pw.is_triggered)
        self.assertFalse(self.pw.trigger_data_done)
        self.assertAlmostEqual(ts, self.pw.trigger_timestamp)

        # Reset the state
        self.pw.is_triggered = False
        self.pw.trigger_data_done = True
        self.pw.trigger_timestamp = 0.0
        self.assertFalse(self.pw.is_triggered)
        self.assertTrue(self.pw.trigger_data_done)
        self.assertAlmostEqual(0.0, self.pw.trigger_timestamp)


        # Test with ai record
        self.pw.trigger_rec_type = "ai"
        self.pw.trigger_level = 10
        data = {}
        data['value'] = 1
        data['timeStamp'] = {}
        data['timeStamp']['secondsPastEpoch'] = 1e+4
        data['timeStamp']['nanoseconds'] = 66
        ts = data['timeStamp']['secondsPastEpoch'] + 1e-9*data['timeStamp']['nanoseconds']

        # Test init state
        self.assertFalse(self.pw.is_triggered)
        self.assertTrue(self.pw.trigger_data_done)
        self.assertAlmostEqual(0.0, self.pw.trigger_timestamp)

        # Run and test state after first, trigger as the first monitor should be
        # ignored because of the init connection call
        self.pw.trigger_process(data)
        self.assertFalse(self.pw.is_triggered)
        self.assertTrue(self.pw.trigger_data_done)
        self.assertAlmostEqual(0.0, self.pw.trigger_timestamp)

        # Call the callback again and check if values were set correctly (values < trigger_level)
        self.pw.trigger_process(data)
        self.assertFalse(self.pw.is_triggered)
        self.assertTrue(self.pw.trigger_data_done)
        self.assertAlmostEqual(0.0, self.pw.trigger_timestamp)

        # Call the callback again and check if values were set correctly (values > trigger_level)
        data['value'] = 11
        self.pw.trigger_process(data)
        self.assertTrue(self.pw.is_triggered)
        self.assertFalse(self.pw.trigger_data_done)
        self.assertAlmostEqual(ts, self.pw.trigger_timestamp)

    def test_is_trigger_in_array(self):
        """

        """
        # Test trigger too small
        self.pw.trigger_timestamp = 1
        time_array = np.arange(100, 200, 0.01)
        in_array, inx = self.pw._PlotWidget__is_trigger_in_array(time_array)
        self.assertFalse(in_array)
        self.assertIsNone(inx)

        # In array
        self.pw.trigger_timestamp = 111
        time_array = np.arange(100, 200, 0.01)
        in_array, inx = self.pw._PlotWidget__is_trigger_in_array(time_array)
        self.assertTrue(in_array)
        self.assertEqual(1100, inx)

        # Test trigger too small
        self.pw.trigger_timestamp = 1111
        time_array = np.arange(100, 200, 0.01)
        in_array, inx = self.pw._PlotWidget__is_trigger_in_array(time_array)
        self.assertFalse(in_array)
        self.assertIsNone(inx)


    def test_data_process(self):
        """
        """
        array_len = 27000
        data = {
            'X_1' : np.random.rand(array_len)*(1e+6) - (1e+6)/2,
            'Y_1' : np.random.rand(array_len)*(1e+6) - (1e+6)/2,
            'X_2' : np.random.rand(array_len)*(1e+6) - (1e+6)/2,
            'Y_2' : np.random.rand(array_len)*(1e+6) - (1e+6)/2,
            'X_3' : np.random.rand(array_len)*(1e+6) - (1e+6)/2,
            'Y_3' : np.random.rand(array_len)*(1e+6) - (1e+6)/2,
            'Time' : np.linspace(100,10_000, num=array_len)
        }

        def data_generator():
            for field, value in data.items():
                yield (field, value)


        # Check init state
        self.assertEqual(0, self.pw.arraysReceived)
        self.assertFalse(self.pw.data) #Empty dict evaluate to False
        self.assertFalse(self.pw.trig_data) #Empty dict evaluate to False
        self.assertEqual(0, self.pw.samples_after_trig_cnt)

        # Call data callback
        self.pw.data_process(data_generator)

        # Check after run (no trigger)
        self.assertEqual(1, self.pw.arraysReceived)
        for field, value in data.items():
             np.testing.assert_array_almost_equal(value[-self.pw.max_length:], self.pw.data[field])
        self.assertFalse(self.pw.trig_data) #Empty dict evaluate to False
        self.assertEqual(0, self.pw.samples_after_trig_cnt)

        # Simulate trigger in the future
        self.pw.trigger_mode = True
        self.pw.is_triggered = True
        self.trigger_timestamp = 11_000

        # Call data callback
        self.pw.data_process(data_generator)

        # Check after run (no trigger)
        self.assertEqual(2, self.pw.arraysReceived)
        for field, value in data.items():
             np.testing.assert_array_almost_equal(np.concatenate((value[-self.pw.max_length:], value)), self.pw.data[field])
        self.assertFalse(self.pw.trig_data) #Empty dict evaluate to False
        self.assertEqual(0, self.pw.samples_after_trig_cnt)

        # Set trigger timestamp in range
        self.pw.trigger_timestamp = 3_766.667

        # Call data callback
        self.pw.data_process(data_generator)

        # Check after run (trigger)
        self.assertEqual(3, self.pw.arraysReceived)
        for field, value in data.items():
             np.testing.assert_array_almost_equal(np.concatenate((value[-2*array_len:], value)), self.pw.data[field])
        self.assertEqual(set(data.keys()), set(self.pw.trig_data.keys()))
        self.assertEqual(17000, self.pw.samples_after_trig_cnt)



class ScopeConfigureBase:
    def __init__(self, params, **kwargs):
        self.params = params
        self.default_arrayid = kwargs.get("arrayid", "None")
        self.default_xaxes = kwargs.get("xaxes", "None")
        self.default_trigger = kwargs.get("trigger", None)
        self.show_trigger = kwargs.get("show_trigger", True)
        self.show_start = kwargs.get("show_start", False)


    def add_source_aquisition_props(self, children, section):
        """
        Add acquisition information specific to the application data source.  This
        should be implemented by the child class
        """
        return children
    
    def assemble_acquisition(self, section=None):
        """
        Assemble acquisition information

        :param section:
        :return:
        """
        
        children = []
        trigger_pv = self.default_trigger
        trigger_mode = False
        trigger_mode_disabled = True
        buffer = None
        
        if section:
            try:
                buffer = section["BUFFER"]
            except ValueError:
                pass

            # Trigger PV name and protocol, which comes with format of proto://pv_name
            # the protocol is either ca:// for channel access or pva:// for pvAccess
            trigger_pv = section.get("TRIGGER", None)
            if self.default_trigger is not None:
                trigger_pv = self.default_trigger

            if trigger_pv is not None and trigger_pv.upper().strip() == "NONE":
                # set trigger PV value to None if a "None" string comes from configuration
                trigger_pv = None
            if trigger_pv is None:
                trigger_mode = None
            else:
                trigger_mode = section.get("WAIT_TRIGGER", None)
            if trigger_mode is not None:
                if trigger_mode.upper().strip() in ["TRUE"]:
                    trigger_mode = True
                else:
                    # set trigger PV value to None if a "None" string comes from configuration
                    trigger_mode = False
            else:
                trigger_mode = False

            trigger_mode_disabled = True
            if trigger_mode:
                trigger_mode_disabled = False

            # post_trigger_pause = 0.0
            # try:
            #     post_trigger_pause = section.getfloat("POST_TRIGGER_PAUSE", 0.0)
            # except ValueError:
            #     pass
            # trigger_holdoff = 0.0
            # try:
            #     trigger_holdoff = section.getfloat("TRIGGER_HOLDOFF", 0.0)
            # except ValueError:
            #     pass

        if self.show_trigger:
            children += [
                {"name": "Trigger PV", "type": "str", "value": trigger_pv},
                {"name": "Trigger Mode", "type": "bool", "value": trigger_mode, "readonly": trigger_mode_disabled},
                # Due to issue: https://github.com/pyqtgraph/pyqtgraph/issues/263
                # always make the trigger level writable instead of disabling writing
                {"name": "Trigger Threshold", "type": "float", "value": 0.0}
                # {"name": "PostTrigger", "type": "float", "value": post_trigger_pause, "siPrefix": True,
                #  "suffix": "Second"},
                # {"name": "HoldTrigger", "type": "float", "value": trigger_holdoff, "siPrefix": True,
                #  "suffix": "Second"},
            ]

        children += [
            {"name": "Freeze", "type": "bool", "value": False},
            {"name": "Buffer (Samples)", "type": "int", "value": buffer, "siPrefix": False, 'decimals': 20}
        ]

        children = self.add_source_acquisition_props(children, section)        

        if self.show_start:
            children.append({"name": "Start", "type": "bool", "value": False})

        acquisition = {"name": "Acquisition", "type":"group", "children": children}
        
        return acquisition

    
    def assemble_display(self, section=None):
        """
        Assemble display information

        :param section:
        :return:
        """
        if section is None:
            display = {"name": "Display", "type": "group", "children": [
                {"name": "Mode", "type": "list", "values": {
                    "Normal": "normal",
                    "FFT": "fft",
                    "PSD": "psd",
                    "Diff": "diff",
                    "X vs Y": "xy",
                }, "value": "Normal"},
                {"name": "FFT filter", "type": "list", "values": {
                    "None" : "none",
                    "Hamming" : "hamming"
                }, "value": "None"},
                {"name": "Exp moving avg", "type": "int", "value": 1, "limits" : (1, 1e+10)},
                {"name": "Autoscale", "type": "bool", "value": False},
                {"name": "Single axis", "type": "bool", "value": True},
                {"name": "Histogram", "type": "bool", "value": False},
                {"name": "Num Bins", "type": "int", "value": 100},
                {"name": "Refresh", "type": "float", "value": 0.1, "siPrefix": True, "suffix": "s"},
            ]}
        else:

            fft_filter = section.get("FFT_FILTER", None)
            if fft_filter is not None:
                if fft_filter.lower().strip() in ["none", "hamming"]:
                    fft_filter = fft_filter.lower().strip().capitalize()
                else:
                    # Disable filter if config not valid
                    fft_filter = "none"

            try:
                n_average = section.getint("AVERAGE", 1)
                n_average = n_average if n_average > 0 else 1
            except ValueError:
                n_average = 1

            autoscale = section.get("AUTOSCALE", None)
            if autoscale is not None:
                if autoscale.upper().strip() in ["TRUE"]:
                    autoscale = True
                else:
                    # set auto scale value to None if a "None" string comes from configuration
                    autoscale = False
            else:
                autoscale = False

            single_axis = section.get("SINGLE_AXIS", None)
            if single_axis is not None:
                if single_axis.upper().strip() in ["FALSE"]:
                    single_axis = False
                else:
                    single_axis = True
            else:
                single_axis = True

            histogram = section.get("HISTOGRAM", None)
            if histogram is not None:
                if histogram.upper().strip() in ["TRUE"]:
                    histogram = True
                else:
                    # set auto scale value to None if a "None" string comes from configuration
                    histogram = False
            else:
                histogram = False

            try:
                n_binning = section.getint("N_BIN", 100)
            except ValueError:
                n_binning = 1

            try:
                refresh = section.getfloat("REFRESH", 100)
            except ValueError:
                refresh = 1

            display = {"name": "Display", "type": "group", "children": [
                {"name": "Mode", "type": "list", "values": {
                    "Normal": "normal",
                    "FFT": "fft",
                    "PSD": "psd",
                    "Diff": "diff",
                    "X vs Y": "xy",
                }, "value": "Normal"},
                {"name": "FFT filter", "type": "list", "values": {
                    "None" : "none",
                    "Hamming" : "hamming"
                }, "value": fft_filter},
                {"name": "Exp moving avg", "type": "int", "value": n_average, "limits" : (1, 1e+10)},
                {"name": "Autoscale", "type": "bool", "value": autoscale},
                {"name": "Single axis", "type": "bool", "value": single_axis},
                {"name": "Histogram", "type": "bool", "value": histogram},
                {"name": "Num Bins", "type": "int", "value": n_binning},
                {"name": "Refresh", "type": "float", "value": refresh / 1000., "siPrefix": True, "suffix": "s"},
            ]}

        return display


    def assemble_config(self):
        # Assemble extra configuration information for plotting
        # which is ArrayId selection, and x axes
        id_value = ["None"]
        if self.default_arrayid != "None":
            id_value.append(self.default_arrayid)
        axes = ["None"]
        if self.default_xaxes != "None":
            axes.append(self.default_xaxes)

        cfg = {"name": "Config",
               "type": "group",
               "children": [
                   {"name": "ArrayId", "type": "list", "values": id_value, "value": self.default_arrayid},
                   {"name": "X Axes", "type": "list", "values": axes, "value": self.default_xaxes},
               ]
               }
        return cfg

    def assemble_statistics(self):
        statistics = {"name": "Statistics", "type": "group", "children": [
            {"name": "CPU", "type": "float", "value": 0, "readonly": True, "suffix": "%"},
            {"name": "Lost Arrays", "type": "int", "value": 0, "readonly": True},
            {"name": "Tot. Arrays", "type": "int", "value": 0, "readonly": True, "siPrefix": True},
            {"name": "Arrays/Sec", "type": "float", "value": 0., "readonly": True, "siPrefix": True,
             "suffix": "/sec"},
            {"name": "Bytes/Sec", "type": "float", "value": 0., "readonly": True, "siPrefix": True,
             "suffix": "/sec"},
            {"name": "Rate", "type": "float", "value": 0., "readonly": True, "siPrefix": True,
             "suffix": "Frames/sec"},
            {"name": "TrigStatus", "type": "str", "value": "", "readonly": True},
        ]}
        return statistics



import unittest
import os
import sys

from configparser import ConfigParser
from pyqtgraph.Qt import QtWidgets
from pyqtgraph.parametertree import Parameter

from c2dataviewer.model import DataSource as DataReceiver
from c2dataviewer.view import StripToolConfigure
from c2dataviewer.striptool import StripToolWindow, WarningDialog, PvEditDialog

from c2dataviewer.control.striptool_config import StriptoolConfig
from c2dataviewer.control.striptool_controller import StripToolController
from c2dataviewer.control.pvconfig import PvConfig

os.environ["QT_QPA_PLATFORM"] = "offscreen"

class TestStriptoolController(unittest.TestCase):
    def setUp(self):
        # Create Qt application
        self.app = QtWidgets.QApplication(sys.argv)

        # Create ImageWindow and get the imageWidget instance
        self.window = StripToolWindow()

        # Build parameter tree
        cf = ConfigParser()
        configure = StripToolConfigure(cf)
        self.parameters = Parameter.create(
            name="params", type="group", children=configure.parse())
        self.window.parameterPane.setParameters(self.parameters, showTop=False)

        # Model to be used
        self.model = DataReceiver()

        # Build GUI elements
        self.warning = WarningDialog(None)
        
        raw = """
[STRIPTOOL]
Chan1.PV = pva://Bar:Baz
Chan2.PV = pva://Bar:Baz2
Chan1.Color = #000000
Chan2.Color = #FFFFFF
        """
        cfg = ConfigParser()
        cfg.read_string(raw)
        self.cfg = cfg

        self.pvedit_dialog = PvEditDialog()
        
        self.scope_controller = StripToolController(
            self.window,  self.model, self.pvedit_dialog, self.warning, self.parameters, self.cfg)

    
    def test_pvlist(self):
        expected = {
            'Bar:Baz' : PvConfig('Bar:Baz', '#000000', 'pva'),
            'Bar:Baz2' : PvConfig('Bar:Baz2', '#FFFFFF', 'pva'),
            }
        pvdict = self.scope_controller.pvdict
        self.assertEqual(len(expected), len(pvdict))
        for k, v in expected.items():
            self.assertTrue(k in pvdict)
            p = pvdict[k].make_pvconfig()
            self.assertEqual(p.pvname, v.pvname)
            self.assertEqual(p.color, v.color)
            self.assertEqual(p.proto, v.proto)


        newpvlist = [ PvConfig('Foo:bar', '#444444', 'ca'), PvConfig('Foo:bar2', '#555555', 'ca') ]
        self.scope_controller.pv_edit_callback(newpvlist)
        self.assertEqual(len(newpvlist), len(pvdict))
        pvdict = self.scope_controller.pvdict
        for v in newpvlist:
            self.assertTrue(v.pvname in pvdict)
            p = pvdict[v.pvname].make_pvconfig()
            self.assertEqual(p.pvname, v.pvname)
            self.assertEqual(p.color, v.color)
            self.assertEqual(p.proto, v.proto)
            
        




# -*- coding: utf-8 -*-

"""
Copyright 2018 UChicago Argonne LLC
 as operator of Argonne National Laboratory

PVA object viewer utilities

@author: Guobao Shen <gshen@anl.gov>
"""

import numpy as np
import pyqtgraph
import pvaccess as pva
from ..model import ConnectionState
from .scope_controller_base import ScopeControllerBase
from ..view.scope_display import PlotChannel as ScopePlotChannel

class ScopeController(ScopeControllerBase):

    def __init__(self, widget, model, parameters, **kwargs):
        
        """

        :param model:
        :param parameters:
        :param channels:
        """
        self.color_pattern = kwargs.get("color", 
                                        ['#FFFF00', '#FF00FF', '#55FF55', '#00FFFF', '#5555FF',
                                         '#5500FF', '#FF5555', '#0000FF', '#FFAA00', '#000000'])

        nchannels = kwargs.get("channels", 4)
        self.channels = []
        for i in range(nchannels):
            self.channels.append(ScopePlotChannel('None', self.color_pattern[i]))

        warning = kwargs["WARNING"]

        super().__init__(widget, model, parameters, warning, channels=self.channels, **kwargs)

        self.auto_buffer_size = self._win.graphicsWidget.max_length is None
            
    def __flatten_dict(dobj, kprefixs=[]):
        """
        Genenerator that can traverse through nested dictionaries and return
        key/value pairs

        For example given {'a':{'b':1}, 'c': 2}, it would yield
        ('a.b', 1) and ('c', 2)

        :param dobj dictionary object
        :param kprefixs  list of key of the directary and it's predecessors
        :yields key, value
        """
        sep = '.'
        for k, v in dobj.items():
            if type(v) == dict:
                yield from ScopeController.__flatten_dict(v, kprefixs + [k])
            else:
                yield sep.join(kprefixs + [k]), v

    def get_fdr(self):
        """
        Get EPICS7 PV field description back as a list

        :return: list of field description
        :raise PvaException: raise pvaccess exception when channel cannot be connected.
        """
        fdr = []
        fdr_scalar = []
        pv = self.model.get()
        
        if pv is None:
            return fdr, fdr_scalar

        pv_structure = pv.getStructureDict()
        pv_dictionary = {k:v for k,v in ScopeController.__flatten_dict(pv_structure)}
        for k, v in pv_dictionary.items():
            if type(v) == list and all(type(e) == pva.ScalarType for e in v):
                # should epics v4 lib not have np, we "fix" it by converting list to np
                v = np.array(v)
                # Make type comparison compatible with PY2 & PY3
                fdr.append(k)
            elif type(v) == pva.ScalarType:
                fdr_scalar.append(k)
            if type(v) != np.ndarray:
                continue
            if len(v) == 0:
                continue

        fdr.sort()
        fdr_scalar.sort()
        return fdr, fdr_scalar

    def update_fdr(self, empty=False):
        """
        Update EPICS7 PV field description

        :return:
        """
        if empty:
            fdr = []
            fdr_scalar = []
        else:
            try:
                fdr, fdr_scalar = self.get_fdr()
            except pva.PvaException as e:
                self.notify_warning('Failed to get PV field description: ' + (str(e)))
                return

        fdr.insert(0, "None")
        fdr_scalar.insert(0, "None")

        # fill up the selectable pull down menu for array ID
        child = self.parameters.child("Config").child("ArrayId")
        child.setLimits(fdr_scalar)
        # fill up the selectable pull down menu for x axes
        child = self.parameters.child("Config").child("X Axes")
        child.setLimits(fdr)

        for idx in range(len(self.channels)):
            child = self.parameters.child("Channel %s" % (idx + 1))
            c = child.child("Field")
            c.setLimits(fdr)
            c.setValue("None")

    def connection_changed(self, state, msg):
        for q in self.parameters.child("Acquisition").children():
            if q.name() == 'PV status':
                q.setValue(state)

    def parameter_change(self, params, changes):
        """

        :param params:
        :param changes:
        :return:
        """
        for param, change, data in changes:
            if change == "value":
                path = self.parameters.childPath(param)
                if path is not None:
                    childName = '.'.join(path)
                else:
                    childName = param.name()

                if childName == "Acquisition.PV":
                    # stop DAQ and update pv info
                    try:
                        self.stop_plotting()
                        self.model.update_device(data, restart=False)
                        if data != "":
                            self.update_fdr()
                        else:
                            self.update_fdr(empty=True)
                        if self.auto_buffer_size:
                            self._win.graphicsWidget.max_length = None
                            
                    except Exception as e:
                        self.notify_warning('Failed to update PV: ' + (str(e)))
                        
                elif childName == "Acquisition.Start":
                    # self.plotting_started = data
                    if data:
                        self.start_plotting()
                    else:
                        self.stop_plotting()
                elif "Channel" in childName:
                    # avoid changes caused by Statistic updating
                    for i, chan in enumerate(self.channels):
                        if childName == 'Channel %s.Field' % (i + 1):
                            chan.pvname = data
                        elif childName == 'Channel %s.DC offset' % (i + 1):
                            chan.dc_offset = data
                        elif childName == 'Channel %s.Axis location' % (i + 1):
                            chan.axis_location = data
                    self._win.graphicsWidget.setup_plot(channels=self.channels)

        super().parameter_change(params, changes)
        
    def monitor_callback(self, data):
        if not self._win.graphicsWidget.max_length:
            for k, v in ScopeController.__flatten_dict(data.get()):
                try:
                    if len(v) > 1:
                        self.update_buffer(len(v))
                        break
                except:
                    pass
            if not self._win.graphicsWidget.max_length:
                return
            
        def generator():
            yield from ScopeController.__flatten_dict(data.get())
        self._win.graphicsWidget.data_process(generator)

    def start_plotting(self):
        """

        :return:
        """
        # stop a model first anyway to ensure it is clean
        self.model.stop()

        # start a new monitor
        self.model.start(self.monitor_callback, self.connection_changed)

        super().start_plotting()
        self.parameters.child("Acquisition").child("Start").setValue(1)

    def stop_plotting(self):
        """

        :return:
        """
        super().stop_plotting()
        self.parameters.child("Acquisition").child("Start").setValue(0)
        # Stop data source
        self.model.stop()

    
